"""
yinter.py

Value interpolator for YAML files.

This project was inspired by the built-in `configparser.ExtendedInterpolation` feature for
`ini` and `cfg` files.
"""
import re
from pathlib import Path

# Dependency packages
import futils
import yaml


class YamlInterpolator(object):
    '''
    Class that consumes a YAML file and returns a dictionary in which values specified in a
    given syntax are interpolated.

    This is heavily based off of the `configparser.ExtendedInterpolation` functionality in
    the `configparser` package.
    '''
    def __init__(self, f_yaml):
        '''
        Initialize the class with a yaml file.

        :param file f_yaml:     The YAML file to process
        '''
        self._f_yaml = Path(f_yaml)

        if not self._f_yaml.exists():
            raise IOError('File does not exist: {}'.format(f_yaml))

        self._d_raw = yaml.load(self._f_yaml.read_text(), Loader=yaml.FullLoader)
        self._d_nested = futils.NestedDict(self._d_raw)
        self._regexpr = r"\$\{([^}]+)\}"

    @property
    def raw_yaml(self):
        '''
        Return a copy of the raw YAML dict. A copy is returned to avoid the chance of modifying
        its contents from the outside.
        '''
        return self._d_raw.copy()

    @property
    def expression(self):
        '''
        Return the regex expression used to parse for interpolated values.
        '''
        return self._regexpr

    def _access_nested_value(self, key, d_node):
        '''
        Access a value nested within the YAML dictionary. The given key represents a path
        to the value by listing relevant keys separated by ':'. For example, accessing 1 in
        { 'a': { 'b': { 'c': 1 } } } would present the key value 'a:b:c' in the interpolation
        string.

        :param str key:         The "key" representing a path to the target value
        :param dict d_node:     The current dictionary node in the YAML tree
        :return obj:            The nested value object
        '''
        t_key_chain = tuple(key.split(':'))
        if len(t_key_chain) == 1:
            return d_node[t_key_chain[0]]
        else:
            return self._d_nested[t_key_chain]

    def _search_and_replace(self, value, d_node):
        '''
        Search a value for interpolation substrings and replace them with the nested value.

        :param str value:       Value (potentially) containing interpolation strings
        :param dict d_node:     The current dictionary node in the YAML tree
        :return None, str:      The interpolated value or None if no interpolation strings are present
        '''
        l_key_paths = re.findall(self._regexpr, value)
        if not l_key_paths:
            return None

        # Set the base interpolated value to the original value
        interpolation = value
        for key in l_key_paths:
            nested_val = self._access_nested_value(key, d_node)
            if nested_val is not None:
                interpolation = interpolation.replace('${' + key + '}', str(nested_val))

        return interpolation

    def _interpolate_string(self, value, d_node):
        '''
        Interpolate a string by checking if it contains interpolation substrings.

        :param str value:       Value (potentially) containing interpolation strings
        :param dict d_node:     The current dictionary node in the YAML tree
        :return obj:            The interpolated or original string
        '''
        interpolated_str = self._search_and_replace(value, d_node)
        return interpolated_str if interpolated_str else value

    def _interpolate_list(self, l_values, d_node):
        '''
        Recursively interpolate a list to find values containing interpolation substrings.
        If the list and its elements contain none, the original list is returned.

        :param list l_values:   The list contents
        :param dict d_node:     The current dictionary node in the YAML tree
        :return list:           A list with one or more interpolated values, or the original list
        '''
        l_interpolated = []
        for value in l_values[:]:
            l_interpolated.append(self._interpolate_for_type(value, d_node))

        return l_interpolated

    def _interpolate_dict(self, d_node):
        '''
        Recursively interpolate a dictionary and its members, interpolating substring values
        as they are found. Dict members are modified directly if they contain interpolation
        substrings.

        :param dict d_node:     The current dictionary node in the YAML tree.
        '''
        for key, value in d_node.copy().items():
            d_node[key] = self._interpolate_for_type(value, d_node)

    def _interpolate_for_type(self, value, d_node):
        '''
        Common function for invoking the proper interpolation procedure for a value's object type.

        :param str value:       Value (potentially) containing interpolation strings
        :param dict d_node:     The current dictionary node in the YAML tree
        '''
        if isinstance(value, dict):
            result = self._interpolate_dict(value)
        elif isinstance(value, list):
            result = self._interpolate_list(value, d_node)
        elif isinstance(value, str):
            result = self._interpolate_string(value, d_node)
        else:
            result = value

        return result if result is not None else value

    def get_interpolated_dict(self):
        '''
        Return a dict representation of a YAML file tree with interpolated values.

        :return dict:   The YAML file as a dict
        '''
        d_interpolated = self._d_nested.copy()
        self._interpolate_dict(d_interpolated)

        return d_interpolated
